try:
    from http.server import HTTPServer, BaseHTTPRequestHandler
except ImportError:
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler

from . import logging
import json

VERSION = 1.1


def get_deep_attr(obj, path):
    if len(path) > 1:
        attr, path = path[0], path[1:]
        return get_deep_attr(getattr(obj, attr), path)
    else:
        return getattr(obj, path[0])


class BoatdHTTPServer(HTTPServer):
    '''
    The main REST server for boatd. Listens for requests on port server_address
    and handles each request with RequestHandlerClass.
    '''
    def __init__(self, boat,
                 server_address, RequestHandlerClass, bind_and_activate=True):

        HTTPServer.__init__(self, server_address, RequestHandlerClass,
                            bind_and_activate)
        self.boat = boat
        self.running = True

        self.handles = {
            '/': self.boatd_info,
            '/boat': self.boat_attr,
            '/wind': self.wind,
            '/active': self.boat_active
        }

        self.post_handles = {
            '/': self.boatd_post,
        }

    def wind(self):
        try:
            speed = self.boat.wind_speed()
        except (AttributeError, TypeError):
            speed = -1

        try:
            return {'direction': self.boat.wind_direction(),
                    'speed': speed}
        except AttributeError as e:
            logging.log(e, logging.ERROR)
            raise AttributeError(e)

    def boat_active(self):
        return {'value': self.boat.active}

    def boatd_info(self):
        return {'boatd': {'version': VERSION}}

    def boat_attr(self):
        return {
            'heading': self.boat.heading(),
            'wind': self.wind(),
            'position': self.boat.position(),
            'active': self.boat.active
        }

    def boatd_post(self, content):
        response = {}
        if 'quit' in content:
            if content.get('quit'):
                self.running = False
                response['quit'] = True

        return response

    def boat_post_function(self, name, content):
        f = self.post_handles.get(name)
        if f is not None:
            return f(content)
        else:
            return self.driver_function(name, args=[content['value']])

    def boat_function(self, function_string):
        '''Return the encoded json response from an endpoint string.'''
        json_content = self.handles.get(function_string)()
        return json_content

    def driver_function(self, function_string, args=None):
        '''
        Return the json response from the string describing the path to the
        attribute.
        '''
        if args is None:
            args = []

        obj_path = [p for p in function_string.split('/') if p]
        attr = get_deep_attr(self.boat, obj_path)
        if callable(attr):
            json_content = {"result": attr(*args)}
        else:
            raise AttributeError
        return json_content


class BoatdRequestHandler(BaseHTTPRequestHandler):
    '''
    Handle a single HTTP request. Returns JSON content using data from the rest
    of boatd.
    '''
    server_version = 'boatd/{}'.format(VERSION)

    def send_json(self, content, code=200):
        self.send_response(code)
        self.send_header('Content-Type', 'application/JSON')
        self.end_headers()
        self.request.sendall(content.encode())

    def do_GET(self, *args, **kwargs):
        '''Handle a GET request to the server.'''
        if self.path in self.server.handles:
            handler_func = self.server.boat_function
        else:
            handler_func = self.server.driver_function

        try:
            func_response = handler_func(self.path)
            code = 200
        except AttributeError:
            func_response = "404 - attribute not found"
            code = 404

        if func_response is not None:
            self.send_json(json.dumps(func_response), code)
        else:
            self.send_json("404", code)

    def do_POST(self):
        '''Handle a POST request to the server.'''
        length = int(self.headers.get('content-length'))
        post_body = self.rfile.read(length).decode('utf-8')
        try:
            data = json.loads(post_body)
        except ValueError:
            logging.log('Can\'t decode {}'.format(post_body), logging.ERROR)
            self.send_json("400 - bad json syntax", 400)
        else:
            response_data = self.server.boat_post_function(self.path, data)
            self.send_json(json.dumps(response_data))

    def log_request(self, code='-', size='-'):
        '''Log the request stdout.'''
        logging.log('{} requested'.format(self.path), level=logging.VERBOSE)



import math

def calculate_checksum(line):
    '''Return the NMEA checksum for a given line'''
    x = 0
    for c in line:
        x ^= ord(c)
    s = str(hex(x))[2:]
    return ('0' if len(s) < 2 else '') + s.upper()

def nmea_line(line):
    checksum = calculate_checksum(line)
    return '$' + line + '*' + checksum

def degrees_to_nmea(input_degrees):
    degrees = math.trunc(input_degrees)
    minutes = (input_degrees - degrees) * 60
    return '{}{:.3g}'.format(degrees, abs(minutes))

def hdm(heading):
    '''Return a HDM nmea sentance from a given heading'''
    line = 'HDM,{0:.3g},M'.format(heading)
    return nmea_line(line)

def gll(latitude, longitude, utc_datetime):
    '''
    Return a GLL nmea sentance from a lat, long and date (datetime object).
    '''
    centisecond = str(utc_datetime.microsecond)[:2]
    t = utc_datetime.strftime('%H%M%S.') + centisecond
    lat_direction = 'N' if latitude > 0 else 'S'
    lon_direction = 'E' if longitude > 0 else 'W'
    line = 'GLL,{lat},N,{lon},W,{time},A'.format(
            lat=degrees_to_nmea(abs(latitude)),
            lon=degrees_to_nmea(abs(longitude)),
            time=t)
    return nmea_line(line)



class Boat(object):
    '''The boat itself. Most of the work is done by the active driver'''
    def __init__(self, driver):
        self.driver = driver
        self.active = False

    def __getattr__(self, name):
        '''Return the requested attribute from the currently loaded driver'''
        return self.driver.handlers.get(name)



import imp
import os
import threading

from . import logging

def get_module_name(filepath):
    _, name = os.path.split(filepath)
    module_name, _ = os.path.splitext(name)
    return module_name

def find_plugins(search_directories):
    found_plugins = []
    for directory in search_directories:
        for plugin in os.listdir(directory):
            if plugin.endswith('.py'):
                found_plugins.append(os.path.join(directory, plugin))

    return found_plugins

def load_plugins(plugin_names):
    modules = []
    for module_filename in plugin_names:
        with open(module_filename) as f:
            module = imp.load_module(
                get_module_name(module_filename),
                f,
                module_filename,
                ('.py', 'U', 1)
            )
            modules.append(module)
            logging.log('loaded plugin from {}'.format(
                                  logging.color(module_filename, 35)))

    return modules

def start_plugins(modules, passed_args):
    for module in modules:
        thread = threading.Thread(target=module.plugin_main,
                                  args=tuple(passed_args))
        thread.start()



from __future__ import print_function

import os
import time

from .color import color

VERBOSE, NORMAL, WARN, ERROR = range(4)


def log(message, level=NORMAL):
    messages = []
    if level > WARN:
        text = 'WARNING'
        if level == ERROR:
            text = 'ERROR'
        messages.append('[{}]'.format(color(text, 31)))

    messages.append(message)
    print(time.strftime('[%H:%M:%S]'), *messages)



from boatd import Driver
driver = Driver()

some_hardware = {}

@driver.heading
def heading():
    return 2.43

@driver.wind_direction
def get_wind():
    return 8.42

@driver.wind_speed
def get_wind_speed():
    return 25

@driver.position
def position():
    return (2.343443, None)

@driver.rudder
def rudder(theta):
    return theta

@driver.sail
def sail(theta):
    return theta



import json


class Config(object):
    '''
    Allow for more natural access to the member values of a dictionary. For
    example, instead of writing:

        d = {
              'a': {
                'x': 5,
                'y': 10
              },
              'b': {
                'z': 42
              }
             }

        d['a']['x']

    we can instead write:

        c = Config(d)
        c.a.x
    '''

    def __init__(self, d):
        self.__dict__.update(d)
        for k, i in self.__dict__.items():
            if isinstance(i, dict):
                self.__dict__[k] = Config(i)

    def get(self, name, default=None):
        if hasattr(self, name):
            return getattr(self, name)
        else:
            return default

    @classmethod
    def from_json(cls, filename):
        '''Return a Config object from a json file'''
        with open(filename) as f:
            return cls(json.load(f))

    @classmethod
    def from_yaml(cls, filename):
        '''Return a Config object from a yaml file'''
        import yaml
        with open(filename) as f:
            return cls(yaml.load(f))



from functools import wraps

from . import logging
from .color import color

class Driver(object):
    def __init__(self):
        self.heading = self.handler('heading')
        self.wind_direction = self.handler('wind_direction')
        self.wind_speed = self.handler('wind_speed')
        self.position = self.handler('position')

        self.rudder = self.handler('rudder',
                function=lambda name, args:
                    logging.log('calling {}({})'.format(name, *args)))

        self.sail = self.handler('sail')

        self.handlers = {}

    def handler(self, name, function=None):
        def wrapper(f):
            @wraps(f)
            def dec(*args, **kwargs):
                if callable(function):
                    function(f.__name__, args)
                return f(*args, **kwargs)
            self.handlers[name] = dec
            logging.log('loaded function {} as {}'.format(
                        color(f.__name__, 32),
                        color('"{}"'.format(name), 35)))
            return dec
        return wrapper



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from subprocess import Popen, PIPE

import boatd

version = Popen(['git', 'describe'],
                stdout=PIPE).communicate()[0].decode('utf8')

if not version.startswith(str(boatd.VERSION)):
    version = boatd.VERSION

print version

setup(
    name='boatd',
    version=version,
    author='Louis Taylor',
    author_email='kragniz@gmail.com',
    description=('Experimental daemon to control an autonomous sailing robot'),
    license='LGPL',
    keywords='boat sailing wrapper rest',
    url='https://github.com/boatd/boatd',
    packages=['boatd'],
    scripts=['boatd-start'],
    requires=['PyYAML'],
    install_requires=[
        'PyYAML >= 3.11'
        ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from __future__ import print_function

import imp
import os
import sys
import traceback

from . import logging
from . import nmea
from . import plugin
from .api import BoatdHTTPServer, BoatdRequestHandler, VERSION
from .boat import Boat
from .color import color
from .config import Config
from .driver import Driver

def load_conf(args):
    '''
    Return the configuration object. Reads from the first argument by default,
    otherwise falls back to 'boatd-config.yaml'.
    '''
    if len(args) > 1:
        conf_file = args[1]
    else:
        conf_file = ''

    _, ext = os.path.splitext(conf_file)
    if ext == '.yaml' or ext == '.yml':
        conf = Config.from_yaml(conf_file)
    elif ext == '.json':
        conf = Config.from_json(conf_file)
    else:
        conf = Config.from_yaml('boatd-config.yaml')

    conf.filename = conf_file

    return conf


def load_driver(conf):
    '''
    Return the driver module from the filename specified in the configuration
    file with key configuration.scripts.driver.
    '''
    expanded_path = os.path.expanduser(conf.scripts.driver)
    directory, name = os.path.split(expanded_path)
    sys.path.append(os.path.dirname(directory))

    if hasattr(conf, 'filename'):
        conf_directory, _ = os.path.split(conf.filename)
        search_dirs = [directory, conf_directory]
    else:
        search_dirs = [directory]

    module_name = os.path.splitext(name)[0]
    try:
        found_module = imp.find_module(module_name, search_dirs)
        driver_module = imp.load_module('driver_module', *found_module)

        _, filename, _ = found_module
        logging.log('loaded driver from {}'.format(
                    color(filename, 34)))

    except Exception as e:
        logging.log('exception raised in driver module:', logging.WARN)
        print(traceback.format_exc())
        raise e
    finally:
        found_module[0].close()

    return driver_module.driver

def load_plugins(conf, boat):
    if conf.get('plugins') is not None:
        plugins = plugin.find_plugins([conf.plugins.directory])
        plugin_modules = plugin.load_plugins(plugins)
        plugin.start_plugins(plugin_modules, [boat])

def run():
    '''Run the main server.'''
    conf = load_conf(sys.argv)
    driver = load_driver(conf)
    boat = Boat(driver)
    load_plugins(conf, boat)

    httpd = BoatdHTTPServer(boat, ('', conf.boatd.port), BoatdRequestHandler)
    while httpd.running:
        try:
            httpd.handle_request()
        except (KeyboardInterrupt, SystemExit):
            logging.log('Quitting...')
            sys.exit()
